"""
A class for constructing Entity classes from EntityModels
"""

from __future__ import annotations

import builtins
from collections.abc import Sequence
from pathlib import Path
from typing import Annotated, Any, Literal

from pydantic import Field, create_model
from pydantic_core import PydanticUndefined
from ruamel.yaml.main import YAML

from ibek.globals import JINJA
from ibek.ibek_builtin.repeat import REPEAT_TYPE, RepeatEntity
from ibek.sub_entity import SubEntity

from .ioc import Entity, EnumVal, clear_entity_model_ids
from .parameters import EnumParam, IdParam, ObjectParam
from .support import EntityModel, Support
from .utils import UTILS


class EntityFactory:
    def __init__(self) -> None:
        """
        A class to create `Entity` types from `EntityModel` instances.

        To understand this class be aware that these are equivalent:-
          EntityModel instance == Entity class.
        i.e. when we instantiate an EntityModel we are creating a new
        dynamic Entity class.

        EntityModel instantiation happens when:-

        1. Deserializing a support module yaml file
        2. calling make_entity_model() in ibek_builtin.repeat.py

        Instantiating an Entity itself is the step that happens when
        deserializing a ioc.yaml. Entities in turn know how to make database
        and startup script entries for IOC instance.

        I have replaced all references in the code to type[Entity] with
        EntityModel for consistency.

        Created models are stored in `self._entity_models` to lookup when
        resolving nested `SubEntity`s.
        """
        self._entity_types: dict[str, EntityModel] = {}
        # starting a new EntityFactory implies we should throw away any existing
        # Entity instances - this is required for tests which create multiple
        # EntityFactories
        clear_entity_model_ids()

    def make_entity_types(self, entity_model_yaml: list[Path]) -> list[EntityModel]:
        """
        Read a set of *.ibek.support.yaml files and generate Entity classes
        from their EntityModel entries
        """

        try:
            for entity_model in entity_model_yaml:
                support_dict = YAML(typ="safe").load(entity_model)

                Support.model_validate(support_dict)

                # deserialize the support module yaml file
                support = Support(**support_dict)
                # make Entity classes described in the support module yaml file
                self._make_entity_types(support)

            # also add builtin entity types "ibek.*"
            self._entity_types[REPEAT_TYPE] = RepeatEntity  # type: ignore
        except Exception:
            print(f"VALIDATION ERROR READING {entity_model}")
            raise

        return list(self._entity_types.values())

    def _make_entity_type(self, model: EntityModel, support: Support) -> EntityModel:
        """
        Create an Entity type from a EntityModel instance and it's containing
        Support instance.
        """

        def add_arg(name, typ, description, default):
            if default is None:
                default = PydanticUndefined
            # cheesy check for enum, can this be improved?
            if typ in [str, object] or "enum" in str(typ):
                args[name] = (Annotated[typ, Field(description=description)], default)
            else:
                args[name] = (
                    Annotated[
                        Annotated[
                            str,
                            Field(
                                description=f"jinja that renders to {typ}",
                                pattern=JINJA,
                            ),
                        ]
                        | Annotated[typ, Field(description=description)],
                        Field(
                            description=f"union of {typ} and jinja "
                            "representation of {typ}"
                        ),
                    ],
                    default,
                )

        args: dict[str, tuple[type, Any]] = {}
        validators: dict[str, Any] = {}

        # fully qualified name of the Entity class including support module
        full_name = f"{support.module}.{model.name}"

        # add in each of the parameters as a Field in the Entity
        for name, param in model.parameters.items():
            type: Any

            if isinstance(param, ObjectParam):
                # we now defer the lookup of the object until whole model validation
                type = object

            elif isinstance(param, IdParam):
                type = str

            elif isinstance(param, EnumParam):
                # Pydantic uses the values of the Enum as the options in the schema.
                # Here we arrange for the keys to be in the schema (what a user supplies)
                # but the values to be what is rendered when jinja refers to the enum
                enum_swapped = {}
                for k, v in param.values.items():
                    enum_swapped[str(v) if v else str(k)] = k
                # TODO review enums especially with respect to Pydantic 2.7.1
                val_enum = EnumVal(name, enum_swapped)  # type: ignore
                type = val_enum

            else:
                # arg.type is str, int, float, etc.
                type = getattr(builtins, param.type)
            add_arg(name, type, param.description, param.default)  # type: ignore

        # add the type literal which discriminates between the different Entity classes
        typ = Literal[full_name]  # type: ignore
        args["type"] = (typ, Field(description=model.description))

        class_name = full_name.replace(".", "_")
        entity_cls = create_model(
            class_name,
            **args,
            __validators__=validators,
            __base__=Entity,
        )  # type: ignore

        # add a link back to the EntityModel Instance that generated this Entity Class
        entity_cls._model = model

        # store this Entity class in the factory
        self._entity_types[full_name] = entity_cls

        return entity_cls

    def _make_entity_types(self, support: Support) -> list[EntityModel]:
        """
        Create Entity subclasses for all EntityModel instances in the given
        Support instance. Returns a list of the Entity subclasses Models.
        """
        entity_names = []
        entity_types = []

        for model in support.entity_models:
            if model.name in entity_names:
                # not tested because schema validation will always catch this first
                raise ValueError(f"Duplicate entity name {model.name}")

            entity_types.append(self._make_entity_type(model, support))

            entity_names.append(model.name)
        return entity_types

    def _resolve_repeat(
        self, repeat_entity: RepeatEntity, context: dict[str, Any]
    ) -> list[Entity]:
        """
        Resolve a repeat Entity into a list of Entity instances
        """
        resolved_entities: list[Entity] = []

        for value in repeat_entity.values:
            context[repeat_entity.variable] = value

            # create the new repeated entity using a dict of arguments
            new_entity = repeat_entity.entity.copy()
            resolved_entities.extend(
                self.resolve_sub_entities([new_entity], context.copy())
            )

        return resolved_entities

    def _make_entity(self, params: dict[str, Any], context: dict[str, Any]) -> Entity:
        # jinja render the parent entity parameters
        for key, param in params.items():
            params[key] = UTILS.render(context, param)

        # create the correct class with new params
        parent_cls = self._entity_types[params["type"]]
        return parent_cls(**params)  # type: ignore

    def resolve_sub_entities(
        self, entities: Sequence[dict | Entity | SubEntity], context: dict[str, Any]
    ) -> list[Entity]:
        """
        Recursively resolve SubEntity collections and Repeat Entities
        in a list of Entity instances

        entities:   list of Entity instances to resolve. These will be Entity
                    subclasses in the root call, but will be dicts in recursive
                    calls via subentities or repeats.
        context:    dictionary of variables to pass to jinja when rendering,
                    this list accumulates as we recurse subentities/repeats
        """
        resolved_entities: list[Entity] = []

        # copy the context for the recursive branches
        context = context.copy()

        for parent_entity in entities:
            if isinstance(parent_entity, dict):
                # convert dictionary to specific Entity with jinja rendering
                parent_entity = self._make_entity(parent_entity, context)

            elif isinstance(parent_entity, SubEntity) or isinstance(
                parent_entity, RepeatEntity
            ):
                # jinja render the arguments of the SubEntity/RepeatEntity
                parent_entity = self._make_entity(parent_entity.model_dump(), context)

            # this parent entity's parameters are passed to children via context
            context.update(parent_entity.model_dump())

            if isinstance(parent_entity, RepeatEntity):
                # resolve repeats in this parent entity
                resolved_entities.extend(self._resolve_repeat(parent_entity, context))
            else:
                # add the current parent entity to the resolved list
                resolved_entities.append(parent_entity)
                # add in SubEntities if any
                for sub_entity in parent_entity._model.sub_entities:
                    # recursively scan the SubEntity for more SubEntities
                    resolved_entities.extend(
                        self.resolve_sub_entities([sub_entity], context)
                    )

        return resolved_entities



from ibek.entity_model import EntityModel
from ibek.ibek_builtin.repeat import RepeatEntity


# Note that an EntityModel instance == Entity class but the type checker
# does not know this. So a few type: ignore comments are needed.
#
# Typically all Entity classes are dynamically generated and treated as
# EntityModel instances. But the builtin entities are defined as classes
# in the ibek_builtin package and offend the type checker.
#
def get_all_builtin_entity_types() -> list[EntityModel]:
    result = [
        RepeatEntity,
    ]
    return result  # type: ignore



"""
System tests for the ibek.repeat feature
"""

from pathlib import Path

from tests.test_cli import generic_generate


def test_repeat_example(tmp_epics_root: Path, tmp_path, samples):
    generic_generate(
        tmp_epics_root,
        samples,
        "repeat",
        ["epics", "asyn", "motorSim"],
    )


def test_subentity_repeats(tmp_epics_root: Path, tmp_path, samples):
    generic_generate(
        tmp_epics_root,
        samples,
        "quadem",
        ["ADCore", "quadem_repeat"],
    )

    # make sure that quadem_repeat and quadem are identical
    files = (samples / "outputs").glob("quadem_repeat/*")
    if len(list(files)) == 0:
        raise AssertionError("No files found in quadem_repeat")

    for f in files:
        assert (samples / "outputs" / "quadem" / f.name).read_text() == f.read_text()


def test_subentity_repeats_with_substitutions(tmp_epics_root: Path, tmp_path, samples):
    generic_generate(
        tmp_epics_root,
        samples,
        "subentity_test",
        ["subentity_test", "epics"],
    )



"""Unit tests for the support subcommand"""

from pathlib import Path

import pytest

from ibek.support_cmds.checks import check_deps, do_dependencies
from ibek.support_cmds.files import symlink_files


def test_symlink_ibek(tmp_path: Path, samples: Path):
    symlink_files(samples / "support", "*.ibek.support.yaml", tmp_path)

    assert sorted([f.name for f in tmp_path.iterdir()]) == [
        "ADCore.ibek.support.yaml",
        "DLS8515.ibek.support.yaml",
        "asyn.ibek.support.yaml",
        "bad_db.ibek.support.yaml",
        "dlsPLC.ibek.support.yaml",
        "epics.ibek.support.yaml",
        "fastVacuum.ibek.support.yaml",
        "gauges.ibek.support.yaml",
        "ipac.ibek.support.yaml",
        "listarg.ibek.support.yaml",
        "motorSim.ibek.support.yaml",
        "quadem.ibek.support.yaml",
        "quadem_repeat.ibek.support.yaml",
        "subentity_test.ibek.support.yaml",
        "technosoft.ibek.support.yaml",
        "utils.ibek.support.yaml",
    ]


def test_symlink_pvi(tmp_path: Path, samples: Path):
    symlink_files(samples / "support", "*.pvi.device.yaml", tmp_path)

    assert [f.name for f in tmp_path.iterdir()] == ["simple.pvi.device.yaml"]


# note this must refer to epics_root to patch where check_deps looks for
# the support files
def test_check_dependencies(tmp_epics_root: Path):
    # Check Passes vs test data
    check_deps(["ADSimDetector", "some-other-module"])

    # Check fails
    with pytest.raises(Exception):  # noqa
        check_deps(["FakeDetector"])


def test_check_release_sh(tmp_epics_root: Path):
    """validate generation of RELEASE.sh for issue #227"""

    do_dependencies()

    release_sh = tmp_epics_root / "support" / "configure" / "RELEASE.shell"
    assert release_sh.exists()
    text = release_sh.read_text()
    assert (
        text
        == """export SUPPORT="/epics/support"
export EPICS_BASE="/epics/epics-base"
export ADSIMDETECTOR="/epics/support/ADSimDetector"
export SOME-OTHER-MODULE="/epics/support/some-other-module"
export EPICS_DB_INCLUDE_PATH=""
"""
    )



"""
Classes for generating an IocInstance derived class from a set of
support module YAML files
"""

from __future__ import annotations

from collections.abc import Sequence
from enum import Enum
from typing import Any

from pydantic import (
    Field,
    model_validator,
)
from pydantic.fields import FieldInfo

from .entity_model import EntityModel
from .globals import BaseSettings
from .utils import UTILS

# a global dict of all entity instances indexed by their ID
id_to_entity: dict[str, Entity] = {}


def get_entity_by_id(id: str) -> Entity:
    try:
        return id_to_entity[id]
    except KeyError as e:
        raise ValueError(f"object {id} not found in {list(id_to_entity)}") from e


def clear_entity_model_ids():
    """Resets the global id_to_entity dict"""

    id_to_entity.clear()


class EnumVal(Enum):
    """
    An enum that is printed as its name only
    """

    def __str__(self):
        return self.name


class Entity(BaseSettings):
    """
    A baseclass for all generated Entity classes.
    """

    type: str = Field(description="The type of this entity")
    entity_enabled: bool = Field(
        description="enable or disable this entity instance", default=True
    )
    _model: EntityModel

    def _process_field(self: Entity, name: str, value: Any, typ: str):
        """
        Process an Entity field - doing jinja rendering, type coercion and
        object id storage/lookup as required.
        """

        if isinstance(value, str):
            # Jinja expansion always performed on string fields
            value = UTILS.render(self, value, typ)

        if typ == "object":
            # look up the actual object by it's id
            if isinstance(value, str):
                value = get_entity_by_id(value)

        # If this field is not pre-existing, add it into the model instance.
        # This is how pre/post_defines are added.
        if name not in self.model_fields:
            self.model_fields[name] = FieldInfo(annotation=str, default=value)

        # update the model instance attribute with the rendered value
        setattr(self, name, value)

        if typ == "id":
            # add this entity to the global id index
            if value in id_to_entity:
                raise ValueError(f"Duplicate id {value} in {list(id_to_entity)}")
            id_to_entity[value] = self

    @model_validator(mode="after")
    def add_ibek_attributes(self):
        """
        Whole Entity model validation

        Do jinja rendering of pre_defines/ parameters / post_defines
        in the correct order.

        Also adds pre_defines and post_defines to the model instance, making
        them available for the phase 2 (final) jinja rendering performed in
        ibek.runtime_cmds.generate().
        """

        # internal 'ibek.' Entity Types do not have _model
        if hasattr(self, "_model"):
            if self._model.pre_defines:
                for name, define in self._model.pre_defines.items():
                    self._process_field(name, define.value, define.type)

            if self._model.parameters:
                for name, parameter in self._model.parameters.items():
                    self._process_field(name, getattr(self, name), parameter.type)

            if self._model.post_defines:
                for name, define in self._model.post_defines.items():
                    self._process_field(name, define.value, define.type)

        return self

    def __str__(self):
        """
        When a jinja template refers to an object by itself e.g.
            # this is the startup entry for {{ my_entity }}
        Jinja will attempt to render the object as a string and this
        method will be called.

        The behaviour is to print the ID of the object. Thus we look up
        which of our object's fields is the ID field and return the
        value of that field.
        """
        id_name = self._model._get_id_arg()
        if id_name:
            return getattr(self, id_name)
        else:
            raise ValueError(
                f"Entity {self.type} has no id field - convert to str failed"
            )

    def __repr__(self):
        return str(self)


class IOC(BaseSettings):
    """
    Used to load an IOC instance entities yaml file into a Pydantic Model.
    """

    ioc_name: str = Field(description="Name of IOC instance")
    description: str = Field(description="Description of what the IOC does")
    entities: list[Entity]
    shared: Sequence[Any] = Field(
        description="A place to create any anchors required for repeating YAML",
        default=(),
    )



"""
Tests for the ibek runtime commands
"""

from ibek.runtime_cmds.autosave import AutosaveGenerator, link_req_files


def test_autosave_generator(samples, tmp_epics_root):
    """Test the autosave generator"""

    link_req_files()
    link_req_files()
    files = list((tmp_epics_root / "autosave").glob("*"))
    assert len(files) == 1

    autosave_gen = AutosaveGenerator(samples / "runtime" / "ioc.subst")
    assert (
        str(autosave_gen.subst_entries[0].template_file)
        == "$(IOCSTATS)/db/iocAdminSoft.db"
    )



"""
A built in entity model instance for generating N instances of another entity.

e.g.
```yaml
  - type: ibek.repeat
    variable: index
    values: {{ range(5) | list }}
    entity:
      type: ADCore.NDPvaPlugin
      NDARRAY_PORT: DET.CAM
      P: BL47P-EA-DET-01
      PORT: DET.PVA{{ index }}
      R: ":TX{{ index }}:"
      PVNAME: BL47P-EA-DET-01:TX{{ index }}:PVA
```
"""

from typing import Literal

from pydantic import Field

from ibek.entity_model import EntityModel
from ibek.ioc import Entity
from ibek.parameters import Param

REPEAT_TYPE = "ibek.repeat"


# TODO this class is an idea to get the schema to work for
# the repeat's entity model.
class EntityModelParam(EntityModel, Param):
    pass


class RepeatEntity(Entity):
    """
    A definition of RepeatEntity for the type checker

    This is not really used - instead the dynamic class is created
    by the make_entity_model function is used.
    """

    type: Literal["ibek.repeat"] = "ibek.repeat"
    values: list | str = Field(
        description="The list of values to iterate over",
    )
    variable: str = Field(
        description="The variable name to use in the entity model",
        default="index",
    )
    entity: dict = Field(
        description="The entity model to repeat",
    )



"""
Some unit tests for ibek.
"""

import dataclasses
import shutil
from pathlib import Path

import pytest

from ibek.commands import semver_compare
from ibek.ioc import id_to_entity
from ibek.ioc_cmds.assets import extract_assets
from ibek.ioc_factory import IocFactory
from ibek.parameters import IdParam, ObjectParam
from ibek.support import EntityModel, Support
from ibek.support_cmds.commands import add_runtime_files
from ibek.utils import UTILS


def test_object_references(entity_factory):
    """
    Verify the object references are correctly resolved
    """
    support = Support(
        module="mymodule",
        entity_models=[
            EntityModel(
                name="port",
                description="a port",
                parameters={"name": IdParam(description="an id")},
            ),
            EntityModel(
                name="device",
                description="a device",
                parameters={"port": ObjectParam(description="the port")},
            ),
        ],
    )

    entities = entity_factory._make_entity_types(support)
    ioc_model = IocFactory().make_ioc_model(entities)
    assert entities[0]._model == support.entity_models[0]
    assert entities[1]._model == support.entity_models[1]

    d = {
        "ioc_name": "",
        "description": "",
        "entities": [
            {"type": "mymodule.port", "name": "PORT"},
            {"type": "mymodule.device", "port": "PORT"},
        ],
    }
    ioc = ioc_model(**d)
    port, device = ioc.entities
    # TODO try to get rid of the need for ''
    assert port.type == "mymodule.port"
    assert device.type == "mymodule.device"
    assert device.port is port
    assert id_to_entity == {"PORT": port}


def test_compare():
    """
    Verify the SemVer comparrisons work
    """
    assert semver_compare("1.1.1", "==1.1.1")
    assert semver_compare("1.1.1", ">=1.1.0")
    assert not semver_compare("1.1.1", ">=1.1.2")


@dataclasses.dataclass
class Person:
    name: str
    age: int


def test_strict():
    """
    validate that bad references in Jinja are caught
    """
    p = Person("giles", 59)

    my_template = "{{ person.name ~ ' of age ' ~ person.age }}"
    text = UTILS.render({"person": p}, my_template)
    assert text == "giles of age 59"

    my_template = "{{ person.name ~ ' of age ' ~ height }}"
    with pytest.raises(ValueError):
        text = UTILS.render({"person": p}, my_template)

    my_template = "{{ person.name ~ ' of age ' ~ person.height }}"
    with pytest.raises(ValueError):
        text = UTILS.render({"person": p}, my_template)


def test_extract_assets(tmp_epics_root: Path, samples: Path):
    """
    Test the extract_assets function
    """
    runtime_files = [
        str(tmp_epics_root / "runtime_file"),
        str(tmp_epics_root / "runtime_folder"),
    ]
    add_runtime_files(runtime_files)

    dest = Path("/tmp/ibek_test_assests")
    shutil.rmtree(dest, ignore_errors=True)
    dest.mkdir()

    extract_assets(dest, tmp_epics_root, [], True)
    new_epics_root = list(dest.glob("tmp/*/*/*/epics"))[0]

    assert Path.exists(new_epics_root / "runtime_file")
    assert Path.exists(new_epics_root / "runtime_folder/text1.txt")
    assert Path.exists(new_epics_root / "runtime_folder/text2.txt")



import os
import subprocess
from pathlib import Path


def expand_env_vars(tokens: list[str]):
    for i in range(len(tokens)):
        tokens[i] = os.path.expandvars(tokens[i])
    return tokens


def handle_command(context: Path, tokens: list[str], step, start):
    msg = f"step {step} {' '.join(tokens)}"
    docker_action = tokens[0]
    tokens = expand_env_vars(tokens)

    print("context,", context)

    if step < start and docker_action != "WORKDIR":
        print("SKIPPING: " + msg)
        return
    else:
        print(msg)

    if docker_action == "RUN":
        result = subprocess.call(["bash", "-c", " ".join(tokens[1:])])
        if result > 0:
            raise RuntimeError("RUN command failed")
    elif docker_action == "WORKDIR":
        folder = Path.cwd() / Path(tokens[1])
        if not folder.exists():
            folder.mkdir(parents=True, exist_ok=True)
        os.chdir(tokens[1])
    elif docker_action == "COPY":
        print("COPY", tokens)
        # skipping because in devcontainer the project folder is already mounted
        # where the destination copy is supposed to go inside the container
    elif docker_action == "FROM":
        if "runtime_prep" in tokens:
            print("\n== Aborting before destructive runtime prep stage. ==\n")
            exit(0)


def build_dockerfile(dockerfile: Path, start: int, stop: int):
    index = 0
    step = 1

    if not dockerfile.exists():
        raise FileNotFoundError(
            "No Dockerfile. Run this command in the generic ioc root folder."
        )
    context = dockerfile.parent

    dockerfile_lines: list[str] = dockerfile.read_text().split("\n")

    stop = min(stop, len(dockerfile_lines))
    while index < len(dockerfile_lines):
        command = dockerfile_lines[index]
        while command.endswith("\\"):
            index += 1
            command = command[:-1] + dockerfile_lines[index]
        index += 1

        if command == "" or command.startswith("#"):
            continue

        handle_command(context, command.split(), step, start)
        step = step + 1
        if step > stop:
            break



"""
Unit tests for the rendering of scripts and database entries from generated
Entity classes
"""

from typing import Literal

from ibek.render import Render
from ibek.render_db import RenderDb


def find_entity_class(entity_classes, type):
    for entity_class in entity_classes:
        literal = Literal[type]  # type: ignore
        if entity_class.model_fields["type"].annotation == literal:
            return entity_class
    else:
        raise ValueError(f"{type} not found in entity_classes")


def test_pre_init_script(asyn_classes):
    definition = find_entity_class(asyn_classes, "asyn.AsynIP")

    # NOTE: because we have switched from type to type in the
    # we need to pass it to the constructor
    # Not ideal - but does not affect deserialization of these types
    port1 = definition(name="asyn1", port="10.0.1.1", type="asyn.AsynIP")
    port2 = definition(name="asyn2", port="10.0.1.2", type="asyn.AsynIP")

    render = Render()
    script_txt = render.render_script(port1, port1._model.pre_init)
    script_txt += render.render_script(port2, port2._model.pre_init)
    assert script_txt == (
        "# Setting up Asyn Port asyn1 on 10.0.1.1:\n"
        "# AsynIPConfigure({{name}}, {{port}}, {{stop}}, {{parity}}, {{bits}}) \n"
        "AsynIPConfigure(asyn1, 10.0.1.1, 1, none, 8)\n"
        "asynSetOption(9600, 0, N, Y)\n"
        'asynOctetSetInputEos("\\n")\n'
        'asynOctetSetOutputEos("\\n")\n'
        "# Setting up Asyn Port asyn2 on 10.0.1.2:\n"
        "AsynIPConfigure(asyn2, 10.0.1.2, 1, none, 8)\n"
        "asynSetOption(9600, 0, N, Y)\n"
        'asynOctetSetInputEos("\\n")\n'
        'asynOctetSetOutputEos("\\n")\n'
    )


def test_obj_ref_script(motor_classes):
    asyn_def = find_entity_class(motor_classes, "asyn.AsynIP")
    motor_def = find_entity_class(motor_classes, "motorSim.simMotorController")

    asyn_def(name="asyn1", port="10.0.1.1", type="asyn.AsynIP")
    motor_obj = motor_def(
        port="asyn1",
        controllerName="ctrl1",
        P="IBEK-MO-01:",
        numAxes=4,
        type="motorSim.simMotorController",
    )

    render = Render()
    script_txt = render.render_script(motor_obj, motor_obj._model.pre_init)

    assert script_txt == (
        "# motorSimCreateController(controller_asyn_port_name, axis_count)\n"
        "# testing escaping:  {{enclosed in escaped curly braces}} \n"
        "motorSimCreateController(ctrl1, 4)\n"
    )


def test_database_render(motor_classes):
    asyn_def = find_entity_class(motor_classes, "asyn.AsynIP")
    sim_def = find_entity_class(motor_classes, "motorSim.simMotorController")
    motor_def = find_entity_class(motor_classes, "motorSim.simMotorAxis")

    asyn1 = asyn_def(name="asyn1", port="10.0.1.1", type="asyn.AsynIP")
    # TODO removing DESC below causes a failure in jinja templating
    # The default DESC field contains jinja escaping and this fails for some
    # reason. But it works fine in the test test_build_runtime_motorSim which
    # is what really counts I guess.
    sim_motor = sim_def(
        port="asyn1",
        controllerName="ctrl1",
        P="IBEK-MO-01:",
        numAxes=4,
        DESC="test",
        type="motorSim.simMotorController",
    )
    motor1 = motor_def(controller="ctrl1", M="M1", ADDR=1, type="motorSim.simMotorAxis")
    motor2 = motor_def(
        controller="ctrl1",
        M="M2",
        ADDR=2,
        is_cs=True,
        type="motorSim.simMotorAxis",
    )

    # make a dummy IOC with two entities as database render works against
    # a whole IOC rather than a single entity at a time.
    entities = [asyn1, sim_motor, motor1, motor2]

    render_db = RenderDb(entities)
    templates = render_db.render_database()

    assert templates == {
        "basic_asyn_motor.db": [
            '"P",           "M",  "DTYP",      "PORT",  "ADDR", "DESC",    "EGU",     "DIR", "VELO", "VMAX", "MRES", "DHLM",  "DLLM",   "INIT"',
            '"IBEK-MO-01:", "M1", "asynMotor", "ctrl1", "1",    "Motor 1", "degrees", "0",   "10.0", "10.0", ".01",  "20000", "-20000", ""    ',
        ],
        "basic_cs_asyn_motor.db": [
            '"P",           "CS_NUM", "DTYP",      "PORT",  "ADDR", "DESC",    "EGU",     "DIR", "VELO", "VMAX", "MRES", "DHLM",  "DLLM",   "INIT"',
            '"IBEK-MO-01:", "0",      "asynMotor", "ctrl1", "2",    "Motor 2", "degrees", "0",   "10.0", "10.0", ".01",  "20000", "-20000", ""    ',
        ],
        "sim_motor.db": [
            '"controllerName", "P",           "DESC"',
            '"ctrl1",          "IBEK-MO-01:", "test"',
        ],
    }


def test_environment_variables(motor_classes):
    asyn_def = find_entity_class(motor_classes, "asyn.AsynIP")

    asyn_obj = asyn_def(name="asyn1", port="10.0.1.1", type="asyn.AsynIP")

    render = Render()
    env_text = render.render_environment_variables(asyn_obj)

    assert env_text == "epicsEnvSet NAME_AS_ENV_VAR my name is asyn1\n"



"""
A class for discovering all autosave req files relating to a given DB
substitution file and generating an equivalent substitution file for
the autosave request files.
"""

import re
from dataclasses import dataclass
from pathlib import Path
from subprocess import run

from ibek.globals import GLOBALS


def link_req_files() -> None:
    """
    Create symbolic links to autosave req files in /epics/autosave folder

    Do this is a fashion that allows overrides:
    1. expect that the folder is already populated with req files from ibek-support
    2. symlink the req files from every built support module's db folder
    2a.  but don't overwrite existing req files
    3. symlink the req files from the instance config folder
    3a.  do overwrite existing req files
    """
    Path.mkdir(GLOBALS.AUTOSAVE, exist_ok=True)
    for req in GLOBALS.SUPPORT.glob("**/*.req"):
        link = GLOBALS.AUTOSAVE / req.name
        if not link.exists():
            link.symlink_to(req)
    for req in (GLOBALS.IOC_FOLDER / GLOBALS.CONFIG_DIR_NAME).glob("*.req"):
        link = GLOBALS.AUTOSAVE / req.name
        if link.exists():
            link.unlink()
        link.symlink_to(req)


class AutosaveGenerator:
    """A class for generating autosave request files for an IOC instance"""

    @dataclass
    class SubstEntry:
        """A dataclass for storing a substitution entry"""

        template_file: Path
        pattern: str

    # https://regex101.com/r/V5cGqt/1
    entry_pat = re.compile(r'file *"?([^" ]*)"[\s\S]*?((?:{[\s\S]*?)*?}[\s\S]})')
    entry_fmt = 'file "{file}" {pattern}\n\n'

    def __init__(self, db_substitution_file: Path):
        """
        Constructor: Build a list of all of the autosave req template files
        related to each of the DB template files in db_substitution_file
        """
        self.db_substitution_file = db_substitution_file
        self.subst_entries = self.parse_subst()
        self.settings_req: dict[str, list[self.SubstEntry]] = {}  # type: ignore
        for suffix in ["settings", "positions"]:
            self.settings_req[suffix] = self.find_req(suffix)

    def generate_req_files(self) -> None:
        """
        Generate a two substitution files for settings and positions req
        template files. Execute MSI to expand these into req files.

        Use these to generate fully expanded req files for the IOC instance.
        These req files are ready to be passed to autosave in the startup script/
        """
        for suffix, entries in self.settings_req.items():
            if not entries:
                continue  # don't generate empty files

            subst_name = GLOBALS.RUNTIME_OUTPUT / f"autosave_{suffix}.subst"
            req_name = GLOBALS.RUNTIME_OUTPUT / f"autosave_{suffix}.req"

            with open(GLOBALS.RUNTIME_OUTPUT / subst_name, "w") as f:
                for entry in entries:
                    short_name = entry.template_file.name
                    f.write(
                        self.entry_fmt.format(file=short_name, pattern=entry.pattern)
                    )
            run(
                f"msi -S {subst_name} > {req_name} -I{GLOBALS.AUTOSAVE}",
                shell=True,
            )

    def parse_subst(self) -> list[SubstEntry]:
        """
        parse an EPICS DB substitution file and return a list of SubstEntry objects
        """
        result = []

        with open(self.db_substitution_file) as f:
            text = f.read()

        match = self.entry_pat.findall(text, re.MULTILINE)
        if match:
            for file, pat in match:
                result.append(self.SubstEntry(template_file=Path(file), pattern=pat))

        return result

    def find_req(self, suffix: str) -> list[SubstEntry]:
        """
        Find autosave req files for each DB template file in
        self.db_substitution_file

        These are files that have the same basename as the DB template file
        with suffix specified in the argument suffix.

        Returns a list of SubstEntry objects with their file modified to match
        the autosave req file name.
        """
        result = []

        for entry in self.subst_entries:
            stem = entry.template_file.stem
            req_file = GLOBALS.AUTOSAVE / f"{stem}_{suffix}.req"
            if req_file.exists():
                result.append(
                    self.SubstEntry(
                        template_file=req_file,
                        pattern=entry.pattern,
                    )
                )

        return result



"""
functions for making Entity models
"""

from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Annotated, Union

from pydantic import Field
from ruamel.yaml.main import YAML

from ibek.utils import UTILS

from .entity_model import EntityModel
from .ioc import IOC


class IocFactory:
    """
    A Class for creating Generic IOC classes and instances
    """

    def deserialize_ioc(
        self, ioc_instance_yaml: Path, entity_models: list[EntityModel]
    ) -> IOC:
        """
        Takes an ioc instance entities file, list of generic ioc yaml files.

        Returns a model of the resulting ioc instance
        """

        # Create a Class representing this Generic IOC
        ioc_model = self.make_ioc_model(entity_models)

        # extract the ioc instance yaml into a dict
        ioc_instance_dict = YAML(typ="safe").load(ioc_instance_yaml)

        if ioc_instance_dict is None or "ioc_name" not in ioc_instance_dict:
            raise RuntimeError(
                f"Failed to load a valid ioc config from {ioc_instance_yaml}"
            )

        # extract the ioc name into UTILS for use in jinja renders
        name = UTILS.render({}, ioc_instance_dict["ioc_name"])
        UTILS.set_ioc_name(name)
        ioc_instance_dict["ioc_name"] = name

        # Create an IOC instance from the instance dict and the model
        ioc_instance = ioc_model(**ioc_instance_dict)

        return ioc_instance

    def make_ioc_model(self, entity_models: list[EntityModel]) -> type[IOC]:
        """
        Create an IOC derived model, by setting its entities attribute to
        be of type 'list of Entity derived classes'.

        Also instantiate any SubEntities at this time
        """

        entity_union = Union[tuple(entity_models)]  # type: ignore
        discriminated = Annotated[entity_union, Field(discriminator="type")]  # type: ignore

        class NewIOC(IOC):
            entities: Sequence[discriminated] = Field(  # type: ignore
                description="List of entities this IOC instantiates"
            )

        return NewIOC



"""
A class containing utility functions for passing into the Jinja context.

This allows us to provide simple functions that can be called inside
Jinja templates with {{ _global.function_name() }}. It also allows
us to maintain state between calls to the Jinja templates because
we pass a single instance of this class into all Jinja contexts.
"""

import ast
import builtins
import os
import re
from collections.abc import Mapping
from pathlib import Path
from typing import Any

from jinja2 import StrictUndefined, Template


class Utils:
    """
    A Utility class for adding functions to the Jinja context
    """

    def __init__(self: "Utils"):
        self.file_name: str = ""
        self.ioc_name: str = ""
        self.__reset__()

    def __reset__(self: "Utils"):
        """
        Reset all saved state. For use in testing where more than one
        IOC is rendered in a single session
        """
        self.variables: dict[str, Any] = {}

    def set_file_name(self: "Utils", file: Path):
        """
        Set the ioc name based on the file name of the instance definition
        """
        self.file_name = file.stem

    def set_ioc_name(self: "Utils", name: str):
        """
        Set the ioc name based on the file name of the instance definition
        """
        self.ioc_name = name

    def get_env(self, key: str) -> str:
        """
        Get an environment variable
        """
        return os.environ.get(key, "")

    def set(self, key: str, value: Any) -> Any:
        """create a global variable for our jinja context"""
        s_key = str(key)
        self.variables[s_key] = value
        return value

    def get(self, key: str, default="") -> Any:
        """get the value a global variable for our jinja context"""
        # default is used to set an initial value if the variable is not set
        s_key = str(key)
        return self.variables.get(s_key, default)

    def incrementor(
        self, name: str, start: int = 0, increment: int = 1, stop: int | None = None
    ) -> int:
        """
        get a named counter that increments by inc each time it is called

        creates a new counter if it does not yet exist
        """
        index = str(name)
        counter = self.variables.get(index)

        if counter is None:
            self.variables[index] = start
        else:
            if not isinstance(counter, int):
                raise ValueError(f"Variable {index} is not an integer")
            self.variables[index] += increment
            if stop is not None and self.variables[index] > stop:
                raise ValueError(f"Counter {index} exceeded maximum value of {stop}")

        return self.variables[index]

    re_get_type = re.compile(r"\| *([a-z]*) *}}")
    typs = ["int", "float", "bool", "list", "dict"]

    def _coerce(self, value: Any, typ: str) -> Any:
        """
        Coerce a value to a given type
        """
        if typ in self.typs:
            if typ in self.typs:
                # make sure the result is the correct type
                cast_type = getattr(builtins, typ)
                return cast_type(ast.literal_eval(value))  # type: ignore
            else:
                raise ValueError(f"Jinja template type '{typ}' not in {self.typs}")
        return value

    def render(self, context: Any, template_text: Any, typ: str = "") -> Any:
        """
        Render a Jinja template with the global _global object in the context
        """
        if isinstance(template_text, list):
            result = (self.render(context, item) for item in template_text)
            result = list(result)  # type: ignore
        elif isinstance(template_text, str):
            # if the template is not a string, jinja render it
            try:
                jinja_template = Template(template_text, undefined=StrictUndefined)
                result = jinja_template.render(  # type: ignore
                    context,
                    # global context for all jinja renders
                    _global=self,
                    # put variables created with set/get directly in the context
                    **self.variables,
                    ioc_yaml_file_name=self.file_name,
                    ioc_name=self.ioc_name,
                )

                if typ:
                    result = self._coerce(result, typ)
                else:
                    match = self.re_get_type.search(template_text)
                    if match:
                        typ = match.group(1)
                        result = self._coerce(result, typ)

            except Exception as e:
                raise ValueError(
                    f"Error rendering template:\n{template_text}\nError: {e}"
                ) from e
        else:
            # because this function is used to template arguments, it may
            # be passed a non string which will always render to itself
            return template_text

        return result

    def render_map(self, context: Any, map: Mapping[str, str | None]) -> dict[str, str]:
        """
        Render a map of jinja templates with values from the given context.

        If given a key with a value of `None`, the key itself will be used as a template
        value, so ``{"P": None}`` is equivalent to ``{"P": "{{ P }}"}``.

        Args:
            context: Context to extract template variables from
            map: Map of macro to jinja template to render

        """
        return {
            key: self.render(
                context,
                template if template is not None else "{{ %s }}" % key,  # noqa
            )
            for key, template in map.items()
        }


# a singleton Utility object for sharing state across all Entity renders
UTILS: Utils = Utils()



import shutil
from pathlib import Path

import typer
from pvi._format.base import IndexEntry
from pvi._format.dls import DLSFormatter
from pvi._format.template import format_template
from pvi.device import Device

from ibek.entity_factory import EntityFactory
from ibek.entity_model import Database
from ibek.gen_scripts import create_boot_script, create_db_script
from ibek.globals import GLOBALS, NaturalOrderGroup
from ibek.ioc import IOC, Entity
from ibek.ioc_factory import IocFactory
from ibek.runtime_cmds.autosave import AutosaveGenerator, link_req_files
from ibek.utils import UTILS

runtime_cli = typer.Typer(cls=NaturalOrderGroup)


@runtime_cli.command()
def generate(
    instance: Path = typer.Argument(
        ...,
        help="The filepath to the ioc instance entity file",
        autocompletion=lambda: [],  # Forces path autocompletion
    ),
    definitions: list[Path] = typer.Argument(
        ...,
        help="The filepath to a support module yaml file",
        autocompletion=lambda: [],  # Forces path autocompletion
    ),
):
    """
    Build a startup script for an IOC instance
    """
    # the file name under of the instance definition provides the IOC name
    UTILS.set_file_name(instance)

    entity_factory = EntityFactory()
    entity_models = entity_factory.make_entity_types(definitions)
    ioc_instance = IocFactory().deserialize_ioc(instance, entity_models)

    # post processing to insert SubEntity instances
    all_entities = entity_factory.resolve_sub_entities(ioc_instance.entities, {})
    ioc_instance.entities = all_entities

    # Clear out generated files so developers know if something stops being generated
    shutil.rmtree(GLOBALS.RUNTIME_OUTPUT, ignore_errors=True)
    GLOBALS.RUNTIME_OUTPUT.mkdir(exist_ok=True)
    shutil.rmtree(GLOBALS.OPI_OUTPUT, ignore_errors=True)
    GLOBALS.OPI_OUTPUT.mkdir(exist_ok=True)

    pvi_index_entries, pvi_databases = generate_pvi(ioc_instance)
    generate_index(ioc_instance.ioc_name, pvi_index_entries)

    script_txt = create_boot_script(ioc_instance.entities)
    script_output = GLOBALS.RUNTIME_OUTPUT / "st.cmd"
    script_output.parent.mkdir(parents=True, exist_ok=True)
    with script_output.open("w") as stream:
        stream.write(script_txt)

    db_txt = create_db_script(ioc_instance.entities, pvi_databases)
    db_output = GLOBALS.RUNTIME_OUTPUT / "ioc.subst"
    with db_output.open("w") as stream:
        stream.write(db_txt)


def generate_pvi(ioc: IOC) -> tuple[list[IndexEntry], list[tuple[Database, Entity]]]:
    """Generate pvi bob and template files to add to UI index and IOC database.

    Args:
        ioc: IOC instance to extract entity pvi definitions from

    Returns:
        List of bob files to add as buttons on index and databases to add to IOC
        substitution file

    """
    index_entries: list[IndexEntry] = []
    databases: list[tuple[Database, Entity]] = []

    formatter = DLSFormatter()

    formatted_pvi_devices: list[str] = []
    for entity in ioc.entities:
        definition = entity._model
        if not hasattr(definition, "pvi") or definition.pvi is None:
            continue
        entity_pvi = definition.pvi

        pvi_yaml = GLOBALS.PVI_DEFS / UTILS.render(entity, entity_pvi.yaml_path)
        device_name = pvi_yaml.name.split(".")[0]
        device_bob = GLOBALS.OPI_OUTPUT / f"{device_name}.pvi.bob"

        # Skip deserializing yaml if not needed
        if entity_pvi.pv or device_name not in formatted_pvi_devices:
            device = Device.deserialize(pvi_yaml)
            device.deserialize_parents([GLOBALS.PVI_DEFS])

            if entity_pvi.pv:
                # Create a template with the V4 structure defining a PVI interface
                output_template = GLOBALS.RUNTIME_OUTPUT / f"{device_name}.pvi.template"
                format_template(device, entity_pvi.pv_prefix, output_template)

                # Add to extra databases to be added into substitution file
                databases.append(
                    (
                        Database(file=output_template.name, args=entity_pvi.ui_macros),
                        entity,
                    )
                )

            if device_name not in formatted_pvi_devices:
                formatter.format(device, device_bob)

                # Don't format further instance of this device
                formatted_pvi_devices.append(device_name)

        if entity_pvi.ui_index:
            macros = UTILS.render_map(dict(entity), entity_pvi.ui_macros)
            index_entries.append(
                IndexEntry(
                    label=f"{device.label}",
                    ui=device_bob.name,
                    macros=macros,
                )
            )

    return index_entries, databases


def generate_index(title: str, index_entries: list[IndexEntry]):
    """Generate an index bob using pvi.

    Args:
        title: Title of index UI
        index_entries: List of entries to include as buttons on index UI

    """
    DLSFormatter().format_index(title, index_entries, GLOBALS.OPI_OUTPUT / "index.bob")


@runtime_cli.command()
def generate_autosave(subst_file: Path = typer.Argument(GLOBALS.RUNTIME_SUBSTITUTION)):
    """
    Generate autosave request files from autosave settings and positions
    template files found in support module db folders. Allow overrides
    from ibek-support/* and /epics/ioc/config folders.

    CONVENTION: req template files must be in the support module db folder after
    the module is built. The template files must be named with the same stem
    as the db file that defines the PVs to be saved. The naming convention is:

    DbTemplateFileStem_positions.req : for autosave stage 0 settings
    DbTemplateFileStem_settings.req : for autosave stage 1 settings

    The steps are:
    1. at build time "ibek support generate-links" creates symlinks to override
       req files in /epics/autosave. These override req files come from individual
       ibek-support folders
    2. at runtime "ibek runtime generate-autosave" handles the remaining steps:
    3. All req files in /epics/support/*/db/ are symlinked to /epics/autosave
       except if that req file name already exists (from 1.)
    4. All req files in /epics/ioc/config are symlinked to /epics/autosave
       overwriting any existing req file (thus supplying instance overrides)
    5. AutosaveGenerator generates two substitution files for settings and positions.
       These substitution files are the same as ioc.db except that they contain
       the file names of the req templates instead of the db files. If ioc.db
       contains a database template that has no corresponding req template, then
       that line is omitted from the new substitution file.
    6. Two req files are generated from the two substitution files using MSI.

    Where the database template has a full path, this will be stripped in
    substitution files created in step 5. At runtime autosave will be pointed at
    the /epics/autosave folder for its search path. There is an issue with name
    collisions that we will address if this ever arises.
    """
    link_req_files()
    asg = AutosaveGenerator(subst_file)
    asg.generate_req_files()



"""
Helper functions for the ibek support commands.
"""

import re
import shutil
from pathlib import Path

import typer

from ibek.globals import GLOBALS, TEMPLATES

# turn RELEASE macros into bash macros
SHELL_FIND = re.compile(r"\$\(([^\)]*)\)")
SHELL_REPLACE = r"${\1}"

# find macros, including ones with blank values
PARSE_MACROS_NULL = re.compile(r"^([A-Z_\-\.a-z0-9]*)\s*=(.*)$", flags=re.M)


def verify_release_includes_local(configure_folder: Path):
    """
    Make sure that a module uses RELEASE.local.

    A git-dirtying patch is required if not.
    """
    validate_support()

    release = configure_folder / "RELEASE"
    text = release.read_text()

    if "RELEASE.local" not in text:
        typer.echo(
            f"WARNING: {configure_folder}/RELEASE does not include RELEASE.local"
        )
        typer.echo(f"WARNING: {configure_folder}/RELEASE will be patched.")
        text += "# PATCHED BY IBEK\n-include $(TOP)/configure/RELEASE.local\n"
        release.write_text(text)


def do_dependencies():
    """
    Fix up the global release file to include all support modules registered
    """
    validate_support()

    # parse the global release file
    global_release_paths = {}
    text = GLOBALS.RELEASE.read_text()
    for match in PARSE_MACROS_NULL.findall(text):
        global_release_paths[match[0]] = match[1]

    # generate the MODULES file for inclusion into the root Makefile
    # it simply defines a variable to hold each of the support module
    # directories in the order they are presented in RELEASE, except that
    s = str(GLOBALS.SUPPORT)
    paths = [
        path[len(s) + 1 :]
        for path in global_release_paths.values()
        if path.startswith(s)
    ]
    if "IOC" in global_release_paths:
        paths.append(global_release_paths["IOC"])
    mod_list = f"MODULES := {' '.join(paths)}\n"
    GLOBALS.MODULES.write_text(mod_list)

    # generate RELEASE.shell file for inclusion into the ioc launch shell script.
    # This adds all module paths to the environment and also adds their db
    # folders to the database search path env variable EPICS_DB_INCLUDE_PATH
    release_sh = []
    for module, path in global_release_paths.items():
        release_sh.append(f'export {module}="{path}"')

    db_paths = [
        f"{path}/db" for path in global_release_paths.values() if path.startswith(s)
    ]
    db_path_list = ":".join(db_paths)
    release_sh.append(f'export EPICS_DB_INCLUDE_PATH="{db_path_list}"')

    shell_text = "\n".join(release_sh) + "\n"
    shell_text = SHELL_FIND.sub(SHELL_REPLACE, shell_text)
    GLOBALS.RELEASE_SH.write_text(shell_text)


def check_deps(deps: list[str]) -> None:
    """
    Check if specified dependencies have been supplied
    """
    for dependency in deps:
        # Check if UCASE module name exist in RELEASE
        with open(GLOBALS.RELEASE) as file:
            release_file = file.read()
            if dependency.upper() in release_file:
                pass
            else:
                raise Exception(f"{dependency.upper()} not in {GLOBALS.RELEASE}")

        # Check if folder with the module name exist in /epics/support
        support_dir = GLOBALS.SUPPORT / dependency
        if Path.exists(support_dir):
            # Check if contains at least one of db, dbd or lib
            res = [Path.exists(support_dir / _dir) for _dir in ["db", "dbd", "lib"]]
            if any(res):
                pass
            else:
                raise Exception(f"db, dbd or lib directory not found in {support_dir}")

        else:
            raise Exception(f"{dependency} directory not in {GLOBALS.SUPPORT}")

        print(f"SUCCESS: {dependency} checked")


def add_macro(macro: str, value: str, file: Path, replace: bool = True):
    """
    add or replace a macro in a RELEASE or CONFIG file
    """
    validate_support()

    if not file.exists():
        file.parent.mkdir(parents=True, exist_ok=True)
        file.touch()

    text = file.read_text()

    find_m = re.compile(f"^({macro}[ \t]*=[ \t]*)(.*)$", flags=re.M)
    matches = find_m.findall(text)
    if len(matches) == 0:
        text += f"{macro}={value}\n"
    elif replace:
        text = find_m.sub(r"\1" + str(value), text)

    file.write_text(text)


def validate_support():
    """
    Validate that the support folder exists and setup initial template files
    if required
    """
    template_support = TEMPLATES / "support"
    release = Path("configure") / "RELEASE"
    global_release = GLOBALS.SUPPORT / release

    if not GLOBALS.SUPPORT.exists():
        typer.echo(f"INITIALIZING {GLOBALS.SUPPORT} folder with template")
        shutil.copytree(template_support, GLOBALS.SUPPORT)
    else:
        if not global_release.exists():
            global_release.parent.mkdir(parents=True, exist_ok=True)
            typer.echo(f"INITIALIZING {GLOBALS.SUPPORT / release} folder with template")
            shutil.copy2(template_support / release, global_release)



"""
Functions for building the db and boot scripts
"""

import logging
from collections.abc import Sequence

from jinja2 import StrictUndefined, Template

from ibek.utils import UTILS

from .entity_model import Database
from .globals import TEMPLATES
from .ioc import Entity
from .render import Render
from .render_db import RenderDb

log = logging.getLogger(__name__)


def create_db_script(
    entities: Sequence[Entity], extra_databases: list[tuple[Database, Entity]]
) -> str:
    """
    Create make_db.sh script for expanding the database templates
    """
    with open(TEMPLATES / "ioc.subst.jinja") as f:
        jinja_txt = f.read()

        renderer = RenderDb(entities)

        templates = renderer.render_database(extra_databases)

        try:
            return Template(jinja_txt).render(
                templates=templates, undefined=StrictUndefined
            )
        except Exception:
            print(f"ERROR RENDERING DATABASE TEMPLATE:\n{templates}")
            raise


def create_boot_script(entities: Sequence[Entity]) -> str:
    """
    Create the boot script for an IOC
    """
    with open(TEMPLATES / "st.cmd.jinja") as f:
        template = Template(f.read())

    renderer = Render()

    return template.render(
        # global context for all jinja renders
        _global=UTILS,
        # put variables created with set/get directly in the context
        **UTILS.variables,
        env_var_elements=renderer.render_environment_variable_elements(entities),
        script_elements=renderer.render_pre_ioc_init_elements(entities),
        post_ioc_init_elements=renderer.render_post_ioc_init_elements(entities),
    )
